
/*
 * Created on 15 avr. 2004
 *
 */

package fr.umlv.sprintabout;

import java.util.concurrent.locks.ReentrantLock;

/**
 * @author remi
 *
 */
public class ConcurrentIntMap extends ReentrantLock {
  
    private static final long serialVersionUID = 6016520323011286353L;
    
    private transient volatile int count;
  private transient HashEntry[] table;
  
  private transient int threshold;
  
  private final float loadFactor;
  
  public ConcurrentIntMap() {
    this(11,0.5f);
  }
  
  public ConcurrentIntMap(int initialCapacity) {
    this(initialCapacity,0.5f);
  }
  
  public ConcurrentIntMap(int initialCapacity, float lf) {
    if (lf<=0.0f || lf>=1.0f)
      throw new IllegalArgumentException("bad load factor "+lf);
    
    int size = 1;
    while (size < initialCapacity) {
      size <<= 1;
    }
    
    this.loadFactor = lf;
    this.table=new HashEntry[size];
    this.threshold = (int)(size * loadFactor);
  }
  
  public int get(Class<?> key) {
    if (count != 0) { // read-volatile
      HashEntry[] tab = table;
      int index = key.hashCode() & (tab.length - 1);
      
      HashEntry e =tab[index];
      while (e != null) {
        if (key==e.key)
          return e.value;
        e = e.next;
      }
    }
    return -1;
  }
  
  public void insert(Class<?> key, int value) {
    lock();
    try {
      int c = count;
      if (c++ > threshold) // ensure capacity
        rehash();
      HashEntry[] tab = table;
      int index = key.hashCode() & (tab.length - 1);
      tab[index] = new HashEntry(key, value, tab[index]);
      count = c; // write-volatile
    } finally {
      unlock();
    }
  }
  
  void rehash() {
    HashEntry[] oldTable = table;            
    int oldCapacity = oldTable.length;
    if (oldCapacity >= MAXIMUM_CAPACITY)
      return;
    
    HashEntry[] newTable = new HashEntry[oldCapacity << 1];
    int sizeMask = newTable.length - 1;
    
    threshold = (int)(newTable.length * loadFactor);
    
    for (int i = 0; i < oldCapacity ; i++) {
      // We need to guarantee that any existing reads of old Map can
      //  proceed. So we cannot yet null out each bin.
      HashEntry e = oldTable[i];
      
      if (e != null) {
        HashEntry next = e.next;
        int idx = e.key.hashCode() & sizeMask;
        
        //  Single node on list
        if (next == null)
          newTable[idx] = e;
        
        else {
          // Reuse trailing consecutive sequence at same slot
          HashEntry lastRun = e;
          int lastIdx = idx;
          for (HashEntry last = next;
          last != null;
          last = last.next) {
            int index = last.key.hashCode() & sizeMask;
            if (index != lastIdx) {
              lastIdx = index;
              lastRun = last;
            }
          }
          newTable[lastIdx] = lastRun;
          
          // Clone all remaining nodes
          for (HashEntry p = e; p != lastRun; p = p.next) {
            int index = p.key.hashCode() & sizeMask;
            newTable[index] = new HashEntry(p.key, p.value,newTable[index]);
          }
        }
      }
    }
    table=newTable;
  }
  
  // all fields must be final to be JSR133 compliant
  private final static class HashEntry {
    final Class<?> key;
    final int value;
    final HashEntry next;
    
    HashEntry(Class<?> key, int value, HashEntry next) {
      this.value = value;
      this.key = key;
      this.next = next;
    }
  }
  
  private static final int MAXIMUM_CAPACITY = 1 << 30; 
}
